import trio
from copy import deepcopy

from asyncowfs.event import (
    ServerRegistered,
    ServerConnected,
    ServerDisconnected,
    ServerDeregistered,
    DeviceAdded,
    DeviceLocated,
    DeviceNotFound,
    BusAdded,
    BusDeleted,
    DeviceValue,
    DeviceAlarm,
)
from asyncowfs.mock import server, EventChecker, structs

import logging

logger = logging.getLogger(__name__)

# We can just use 'async def test_*' to define async tests.
# This also uses a virtual clock fixture, so time passes quickly and
# predictably.

basic_tree = {
    "bus.0": {
        "alarm": {},
        "simultaneous": {"temperature": 0},
        "10.345678.90": {"latesttemp": "12.5", "templow": "15", "temphigh": "20"},
    },
    "structure": structs,
}


async def test_alarm(mock_clock):
    e1 = EventChecker(  # pylint: disable=unused-variable
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            DeviceAlarm("10.345678.90"),
            DeviceValue("10.345678.90", "temperature", 12.5),
            # DeviceValue("10.345678.90", "temperature",12.5),
            # DeviceNotFound("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    mock_clock.autojump_threshold = 0.1
    my_tree = deepcopy(basic_tree)
    dt = my_tree["bus.0"].pop("10.345678.90")
    async with server(tree=my_tree) as ow:  # , polling=True, events=e1) as ow:
        dev = await ow.get_device("10.345678.90")
        dev.interval_alarm = 10
        dev.interval_temperature = 15

        my_tree["bus.0"]["10.345678.90"] = dt
        await ow.scan_now()
        my_tree["bus.0"]["alarm"]["10.345678.90"] = dt
        await trio.sleep(12)  # allow alarm poll to trigger
        assert int(dt["temphigh"]) == 20
        assert int(dt["templow"]) == 11
        assert dev.alarm_temperature == 12.5
        await trio.sleep(5)  # allow temperature poll to trigger



import trio
import pytest
from copy import deepcopy

from moat.util import ungroup,exc_iter

import anyio

from asyncowfs.error import NoEntryError
from asyncowfs.event import (
    ServerRegistered,
    ServerConnected,
    ServerDisconnected,
    ServerDeregistered,
    DeviceAdded,
    DeviceLocated,
    DeviceNotFound,
    BusAdded_Path,
    BusAdded,
    BusDeleted,
)
from asyncowfs.bus import Bus

from asyncowfs.mock import server, EventChecker, structs

import logging

logger = logging.getLogger(__name__)

# We can just use 'async def test_*' to define async tests.
# This also uses a virtual clock fixture, so time passes quickly and
# predictably.

basic_tree = {
    "bus.0": {
        "10.345678.90": {
            "whatever": "hello",
            "latesttemp": "12.5",
            "temperature": "12.5",
            "templow": "10",
            "temphigh": "20",
            "foo": {
                "bar": "42",
                "baz": {"quux": "99.875"},
                "plugh.A": 1,
                "plugh.B": 2,
                "plugh.C": 3,
                "plover.0": 7,
                "plover.1": 8,
                "plover.2": 9,
            },
        }
    },
    "structure": structs,
}

coupler_tree = {
    "bus.0": {
        "10.345678.90": {"whatever": "hello", "temperature": "12.5"},
        "1F.ABCDEF.F1": {
            "main": {"20.222222.22": {"some": "chip"}},
            "aux": {"28.282828.28": {"another": "chip"}},
        },
    },
    "structure": structs,
}


async def test_empty_server():
    e1 = EventChecker([ServerRegistered, ServerConnected, ServerDisconnected, ServerDeregistered])
    async with server(events=e1):  # as ow:
        await trio.sleep(0)


async def test_send_idle(mock_clock):
    mock_clock.autojump_threshold = 0.1
    e1 = EventChecker([ServerRegistered, ServerConnected, ServerDisconnected, ServerDeregistered])
    async with server(events=e1):  # as ow:
        await trio.sleep(15)


async def test_missing_event():
    e1 = EventChecker([ServerRegistered, ServerConnected, ServerDisconnected])
    with pytest.raises((RuntimeError,ExceptionGroup)) as r, ungroup:
        async with server(events=e1):  # as ow:
            await trio.sleep(0)

    for e in exc_iter(r.value):
        if isinstance(e, RuntimeError):
            assert "Unexpected event " in e.args[0]
        elif isinstance(e, anyio.ClosedResourceError):
            pass
        else:
            raise


async def test_more_event():
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerDeregistered,
            ServerDeregistered,
        ]
    )
    with pytest.raises(RuntimeError) as r, ungroup:
        async with server(events=e1):  # as ow:
            await trio.sleep(0)
    assert "Superfluous event " in r.value.args[0]


async def test_bad_event():
    e1 = EventChecker([ServerRegistered, ServerConnected, ServerConnected, ServerDeregistered])
    with pytest.raises((RuntimeError,ExceptionGroup)) as r:
        async with server(events=e1):  # as ow:
            await trio.sleep(0)
    for e in exc_iter(r.value):
        if isinstance(e, RuntimeError):
            assert "Wrong event: want " in e.args[0]
        elif isinstance(e, anyio.ClosedResourceError):
            pass
        else:
            raise




async def test_basic_server():
    e1 = EventChecker(  # pylint: disable=unused-variable
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus.0"),
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound,
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(tree=basic_tree) as ow:  # events=e1) as ow:
        await trio.sleep(0)
        dev = await ow.get_device("10.345678.90")
        assert dev.bus == ("bus.0",)
        assert isinstance(dev.bus, Bus)
        assert float(await dev.attr_get("temperature")) == 12.5
        await dev.attr_set("temperature", value=98.25)
        assert float(await dev.attr_get("temperature")) == 98.25
        await dev.attr_set("temperature", value=12.5)


async def test_basic_structs(mock_clock):
    mock_clock.autojump_threshold = 0.1
    async with server(
        tree=basic_tree,
        options={"slow_every": [0, 2, 0.1], "busy_every": [0, 0, 1], "close_every": [0, 0, 0, 1]},
    ) as ow:
        await trio.sleep(0)
        dev = await ow.get_device("10.345678.90")
        await ow.ensure_struct(dev)
        await dev.wait_bus()
        assert await dev.temperature == 12.5
        await dev.set_temphigh(98.25)
        assert await dev.temphigh == 98.25

        # while we're at it, test our ability to do things in parallel on a broken server
        dat = {}
        evt = trio.Event()

        async def get_val(tag):
            await evt.wait()
            dat[tag] = await getattr(dev, tag)

        async with trio.open_nursery() as n:
            n.start_soon(get_val, "temperature")
            n.start_soon(get_val, "temphigh")
            n.start_soon(get_val, "templow")
            await trio.sleep(1)
            evt.set()
        assert dat == {"temphigh": 98.25, "temperature": 12.5, "templow": 10.0}, dat


async def test_more_structs(mock_clock):
    mock_clock.autojump_threshold = 0.1
    async with server(tree=basic_tree) as ow:
        await trio.sleep(0)
        dev = await ow.get_device("10.345678.90")
        await ow.ensure_struct(dev)
        assert await dev.temperature == 12.5
        await dev.set_temphigh(98.25)
        assert await dev.temphigh == 98.25
        assert await dev.foo.get_bar() == 42
        assert await dev.foo.baz.quux == 99.875
        assert await dev.foo.plugh[0] == 1
        assert await dev.foo.plugh[1] == 2
        assert await dev.foo.plugh[2] == 3
        assert await dev.foo.plover[0] == 7
        assert await dev.foo.get_plover(1) == 8
        assert await dev.foo.plover[2] == 9
        await dev.foo.set_bar(123)
        await dev.foo.baz.set_quux(234)
        await dev.foo.set_plugh(1, 11)
        await dev.foo.set_plover(1, 111)
        assert await dev.foo.bar == 123
        assert await dev.foo.baz.quux == 234
        assert await dev.foo.plugh[0] == 1
        assert await dev.foo.plugh[1] == 11
        assert await dev.foo.plugh[2] == 3
        assert await dev.foo.plover[0] == 7
        assert await dev.foo.plover[1] == 111
        assert await dev.foo.get_plover(2) == 9


async def test_coupler_server(mock_clock):
    mock_clock.autojump_threshold = 0.1
    e1 = EventChecker(  # pylint: disable=unused-variable
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus.0"),
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            DeviceAdded("1F.ABCDEF.F1"),
            DeviceLocated("1F.ABCDEF.F1"),
            BusAdded_Path("bus.0", "1F.ABCDEF.F1", "main"),
            DeviceAdded("20.222222.22"),
            DeviceLocated("20.222222.22"),
            BusAdded_Path("bus.0", "1F.ABCDEF.F1", "aux"),
            DeviceAdded("28.282828.28"),
            DeviceLocated("28.282828.28"),
            ServerDisconnected,
            DeviceNotFound,
            BusDeleted,
            DeviceNotFound,
            BusDeleted,
            DeviceNotFound,
            DeviceNotFound,
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(tree=coupler_tree):  # events=e1, tree=coupler_tree):  # as ow:
        await trio.sleep(0)


async def test_wrong_bus():
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus.0"),
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            ServerDeregistered,
        ]
    )
    with pytest.raises((RuntimeError,ExceptionGroup)) as r:
        async with server(events=e1, tree=basic_tree):  # as ow:
            await trio.sleep(0)
    for e in exc_iter(r.value):
        if isinstance(e, RuntimeError):
            assert "Wrong event: want " in e.args[0]
        elif isinstance(e, anyio.ClosedResourceError):
            pass
        else:
            raise




async def test_slow_server(mock_clock):
    mock_clock.autojump_threshold = 0.1
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(
        events=e1, tree=basic_tree, options={"slow_every": [0, 0, 5, 0, 0]}
    ):  # as ow:
        await trio.sleep(0)


async def test_busy_server():
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(events=e1, tree=basic_tree, options={"busy_every": [0, 0, 1]}):
        await trio.sleep(0)
    async with server(events=e1, tree=basic_tree, options={"busy_every": [1, 0, 1]}):
        await trio.sleep(0)


async def test_disconnecting_server():
    e1 = EventChecker(  # pylint: disable=unused-variable
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            BusAdded,
            ServerDisconnected,
            ServerConnected,
            DeviceAdded("10.345678.90"),
            ServerDisconnected,
            ServerConnected,
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(
        tree=basic_tree, options={"close_every": [0, 0, 0, 1]}  # events=e1,
    ):  # as ow:
        await trio.sleep(0)


async def test_disconnecting_server_2(mock_clock):
    mock_clock.autojump_threshold = 0.1
    e1 = EventChecker(  # pylint: disable=unused-variable
        [
            ServerRegistered,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            ServerDisconnected,
            ServerConnected,
            BusAdded,
            ServerDisconnected,
            ServerConnected,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(
        tree=basic_tree, options={"close_every": [0, 0, 1, 0, 0]}  # events=e1, tree=basic_tree,
    ):  # as ow:
        await trio.sleep(0)


async def test_dropped_device(mock_clock):
    mock_clock.autojump_threshold = 0.1
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            DeviceNotFound("10.345678.90"),
            ServerDisconnected,
            BusDeleted,
            ServerDeregistered,
        ]
    )
    my_tree = deepcopy(basic_tree)
    async with server(tree=my_tree, events=e1) as ow:
        dev = await ow.get_device("10.345678.90")
        assert dev.bus is not None
        del my_tree["bus.0"]["10.345678.90"]
        assert dev._unseen == 0

        await ow.scan_now(polling=False)
        assert dev._unseen == 1
        await ow.scan_now(polling=False)
        assert dev._unseen == 2
        await ow.scan_now(polling=False)
        assert dev._unseen == 3
        assert dev.bus is not None
        await ow.scan_now(polling=False)
        assert dev._unseen == 3
        assert dev.bus is None


async def test_manual_device(mock_clock):
    class Checkpoint:
        pass

    mock_clock.autojump_threshold = 0.1
    e1 = EventChecker(
        [
            ServerRegistered,
            ServerConnected,
            BusAdded,
            Checkpoint,
            DeviceAdded("10.345678.90"),
            Checkpoint,
            DeviceLocated("10.345678.90"),
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    my_tree = deepcopy(basic_tree)
    entry = my_tree["bus.0"].pop("10.345678.90")
    async with server(events=e1, tree=my_tree) as ow:
        await ow.push_event(Checkpoint())
        dev = await ow.get_device("10.345678.90")
        assert dev.bus is None
        await trio.sleep(15)
        await ow.push_event(Checkpoint())
        my_tree["bus.0"]["10.345678.90"] = entry
        await ow.scan_now(polling=False)
        assert dev.bus is not None


async def test_nonexistent_device(mock_clock):
    mock_clock.autojump_threshold = 0.1
    my_tree = deepcopy(basic_tree)
    my_tree.pop("bus.0")
    async with server(tree=my_tree, scan=None) as ow:
        bus = await ow.test_server.get_bus("bus.0")
        assert bus.server == ow.test_server

        dev = await ow.get_device("10.345678.90")
        await dev.locate(bus)
        with pytest.raises(NoEntryError):
            assert float(await dev.attr_get("temperature")) == 12.5


async def test_manual_bus(mock_clock):
    class Checkpoint:
        pass

    mock_clock.autojump_threshold = 0.1
    e1 = EventChecker(  # pylint: disable=unused-variable
        [
            ServerRegistered,
            ServerConnected,
            BusAdded_Path("bus.0"),
            DeviceAdded("10.345678.90"),
            DeviceLocated("10.345678.90"),
            Checkpoint,
            Checkpoint,
            ServerDisconnected,
            DeviceNotFound("10.345678.90"),
            BusDeleted,
            ServerDeregistered,
        ]
    )
    async with server(tree=basic_tree, scan=None, initial_scan=False) as ow:  # events=e1,
        bus = await ow.test_server.get_bus("bus.0")
        assert bus.server == ow.test_server

        dev = await ow.get_device("10.345678.90")
        assert dev.bus is None
        await dev.locate(bus)
        assert dev.bus is not None
        assert float(await dev.attr_get("temperature")) == 12.5



"""
Access to an owserver.
"""

import anyio

from collections import deque
from typing import Union
from functools import partial
from concurrent.futures import CancelledError

from .event import ServerConnected, ServerDisconnected
from .event import BusAdded
from .protocol import (
    NOPMsg,
    DirMsg,
    AttrGetMsg,
    AttrSetMsg,
    MessageProtocol,
    ServerBusy,
)
from .bus import Bus
from .util import ValueEvent

import logging

logger = logging.getLogger(__name__)


class Server:
    """\
        Encapsulate one server connection.
    """

    def __init__(self, service, host="localhost", port=4304, name=None):
        self.service = service
        self.host = host
        self.port = port
        self.name = name or host
        self.stream = None
        self._msg_proto = None
        self.requests = deque()
        self._wqueue_w, self._wqueue_r = anyio.create_memory_object_stream(100)
        self._read_task = None
        self._write_task = None
        self._scan_task = None
        self._buses = dict()  # path => bus
        self._scan_lock = anyio.Lock()
        self._scan_args = None
        self._backoff = 2
        self._current_tg = None
        self._current_run = None

        self.simul_lock = anyio.Lock()

    async def get_bus(self, *path):
        """Return the bus at this path. Allocate new if not existing."""
        try:
            return self._buses[path]
        except KeyError:
            bus = Bus(self, *path)
            self._buses[bus.path] = bus
            await self.service.push_event(BusAdded(bus))
            return bus

    def __repr__(self):
        return "<%s:%s:%d %s>" % (
            self.__class__.__name__,
            self.host,
            self.port,
            "OK" if self.stream else "closed",
        )

    async def _reader(self, evt):
        try:
            with anyio.CancelScope() as scope:
                self._read_task = scope
                evt.set()
                it = self._msg_proto.__aiter__()
                while True:
                    try:
                        with anyio.fail_after(15):
                            res, data = await it.__anext__()
                    except StopAsyncIteration:
                        raise anyio.ClosedResourceError from None
                    except ServerBusy:
                        logger.debug("Server %s busy", self.host)
                    else:
                        msg = self.requests.popleft()
                        await msg.process_reply(res, data, self)
                        if not msg.done():
                            self.requests.appendleft(msg)
        except anyio.ClosedResourceError:
            if self._current_tg is not None:
                self._current_tg.cancel_scope.cancel()

    async def _run_one(self, val: ValueEvent):
        try:
            async with anyio.create_task_group() as tg:
                self._current_tg = tg
                self.stream = await anyio.connect_tcp(self.host, self.port)

                ml, self.requests = deque(self.requests), deque()
                try:
                    self._msg_proto = MessageProtocol(self, is_server=False)

                    e_w = anyio.Event()
                    e_r = anyio.Event()
                    tg.start_soon(self._writer, e_w)
                    tg.start_soon(self._reader, e_r)
                    await e_r.wait()
                    await e_w.wait()

                    # re-send messages, but skip those that have been cancelled
                    while ml:
                        msg = ml.popleft()
                        if not msg.cancelled:
                            try:
                                await self._wqueue_w.send(msg)
                            except BaseException:
                                ml.appendleft(msg)

                except BaseException:
                    self.requests = ml
                    raise

                await self.chat(NOPMsg())

                if self._scan_args is not None:
                    tg.start_soon(partial(self.start_scan, **self._scan_args))
                if val is not None:
                    val.set(None)
                self._backoff = 0.1
                pass  # wait for tasks
            pass  # exited tasks

        finally:
            self._current_tg = None
            if self.stream is not None:
                with anyio.CancelScope(shield=True):
                    await self.stream.aclose()
                self.stream = None

    async def start(self):
        """Start talking. Returns when the connection is established,
        raises an error if that's not possible.
        """
        val = ValueEvent()
        self.service.nursery.start_soon(self._run_reconnected, val)
        await val.get()
        await self.service.push_event(ServerConnected(self))

    async def _run_reconnected(self, val: ValueEvent):
        try:
            with anyio.CancelScope() as scope:
                self._current_run = scope
                while True:
                    try:
                        await self._run_one(val)
                    except anyio.get_cancelled_exc_class():
                        raise
                    except (
                        BrokenPipeError,
                        TimeoutError,
                        EnvironmentError,
                        anyio.IncompleteRead,
                        ConnectionResetError,
                        anyio.ClosedResourceError,
                        StopAsyncIteration,
                    ) as exc:
                        if val is not None and not val.is_set():
                            val.set_error(exc)
                            return
                        logger.error("Disconnected")
                        val = None

                        await anyio.sleep(self._backoff)
                        if self._backoff < 10:
                            self._backoff *= 1.5
                    else:
                        pass
        finally:
            self._current_run = None

    async def setup_struct(self, dev):
        await dev.setup_struct(self)

    async def chat(self, msg):
        await self._wqueue_w.send(msg)
        try:
            res = await msg.get_reply()
            return res
        except BaseException:
            msg.cancel()
            raise

    async def _writer(self, evt):
        with anyio.CancelScope() as scope:
            self._write_task = scope
            evt.set()
            while True:
                msg = None
                try:
                    logger.debug("Writer Start %s", self.host)
                    with anyio.fail_after(10):
                        msg = await self._wqueue_r.receive()
                        logger.debug("Writer send %s", msg)
                except TimeoutError:
                    if msg is None:
                        msg = NOPMsg()
                        logger.debug("Writer TO send %s", msg)
                    else:
                        logger.warning("Writer OWCHIE send")

                self.requests.append(msg)
                await msg.write(self._msg_proto)
                logger.debug("Writer did %s", msg)

    async def drop(self):
        """Stop talking and delete yourself"""
        try:
            await self.aclose()
        finally:
            await self.service._del_server(self)

    async def aclose(self):
        if self._current_run is not None:
            self._current_run.cancel()
        if self._current_tg is not None:
            self._current_tg.cancel_scope.cancel()

        await self.service.push_event(ServerDisconnected(self))

        if self._buses is not None:
            for b in list(self._buses.values()):
                await b.delocate()
        self._buses = None
        for m in self.requests:
            m.cancel()

    @property
    def all_buses(self):
        for b in list(self._buses.values()):
            yield from b.all_buses

    async def dir(self, *path):
        return await self.chat(DirMsg(path))

    async def _scan(self, interval, initial_interval, polling, random=0):
        if not initial_interval:
            initial_interval = interval
        # 5% variation, to prevent clustering
        if random:
            initial_interval *= 1 + (random() - 0.5) / random
        await anyio.sleep(initial_interval)

        while True:
            async with self._scan_lock:
                await self.scan_now(polling=polling)
            if not interval:
                return
            i = interval
            if random:
                i *= 1 + (random() - 0.5) / random
            await anyio.sleep(i)

    async def scan_now(self, polling=True):
        if self._scan_lock.locked():
            # scan in progress: just wait for it to finish
            async with self._scan_lock:
                pass
        else:
            async with self._scan_lock:
                await self._scan_base(polling=polling)

    async def _scan_base(self, polling=True):
        old_paths = set()

        # step 1: enumerate
        try:
            for d in await self.dir():
                if d.startswith("bus."):
                    bus = await self.get_bus(d)
                    bus._unseen = 0
                    try:
                        old_paths.remove(d)
                    except KeyError:
                        pass
                    buses = await bus._scan_one(polling=polling)
                    old_paths -= buses
        except CancelledError:
            return

        # step 2: deregister buses, if not seen often enough
        for p in old_paths:
            bus = self._buses.get(p, None)
            if bus is None:
                continue
            if bus._unseen > 2:
                await bus.delocate()
            else:
                bus._unseen += 1

    async def start_scan(
        self,
        scan: Union[float, None] = None,
        initial_scan: Union[float, bool] = True,
        polling=True,
        random: int = 0,
    ):
        """Scan this server.

        :param scan: Flag how often to re-scan the bus.
            None: don't scan at all
            >0: repeat in the background
        :param initial_scan: Flag when to initially scan the bus.
            False: don't.
            True: immediately, wait until complete.
            >0: return immediately, delay initial scan that many seconds.
        :type scan: :class:`float` or ``None``
        :type initial_scan: :class:`float` or :class:`bool`
        :param polling: Flag whether to start tasks for periodic polling
            (alarm handling, temperature, …). Defaults to ``True``.
        """
        self._scan_args = dict(scan=scan, initial_scan=False, polling=polling, random=random)
        if not scan and not initial_scan:
            return
        if scan and scan < 1:
            raise RuntimeError("You can't scan that often.")
        if initial_scan is True:
            await self.scan_now(polling=polling)
            initial_scan = False

        if initial_scan or scan:
            self._scan_task = await self._current_tg.start_soon(
                self._scan, scan, initial_scan, polling, random
            )

    async def attr_get(self, *path):
        return await self.chat(AttrGetMsg(*path))

    async def attr_set(self, *path, value):
        return await self.chat(AttrSetMsg(*path, value=value))



import os
import trio
import anyio
from asyncowfs import OWFS
from asyncowfs.protocol import MessageProtocol, OWMsg
from asyncowfs.error import OWFSReplyError, NoEntryError, IsDirError

from moat.util import ungroup

try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager
from functools import partial

import logging

logger = logging.getLogger(__name__)

# pylint: disable=raise-missing-from


def _chk(v):
    if v is None:
        return False
    v[0] += 1
    if v[0] >= len(v):
        v[0] = 1
    return v[v[0]]


async def _schk(v):
    if v is None:
        return
    v[0] += 1
    if v[0] >= len(v):
        v[0] = 1
    if v[v[0]] > 0:
        logger.debug("Slow reply %s", v[v[0]])
    await trio.sleep(v[v[0]])


class FakeMaster:
    def __init__(self, stream):
        self.stream = stream


async def some_server(tree, options, socket):
    """
    This is a fake 1wire server task.

    ``tree``: the 1wire hierarchy to serve::

        {
            "bus.0": {
                "10.345678.90": {
                    "latesttemp": "12.5",
                    "temperature": "12.5",
                    "templow": "10",
                    "temphigh": "20",
                }
            },
            "structure": asyncowfs.mock.structs,
        }


    ``options``: an object with optional ``busy_every``, ``close_every``
    and/or ``slow_every`` attributes. These must be an array with an offset
    and a list of flags. Each call cycles through the array and,
    respectively, reports to be busy, closes the connection, or delays its
    answer. This allows you to test various interesting bus conditions.
    See the ``tests/test_example.pytest_basic_structs`` test for example
    use.

    ``socket``: the connection to the client.
    """
    rdr = MessageProtocol(FakeMaster(socket), is_server=True)
    logger.debug("START Server")

    each_busy = options.get("busy_every", None)
    each_close = options.get("close_every", None)
    each_slow = options.get("slow_every", None)
    try:
        if _chk(each_close):
            return
        async for command, format_flags, data, offset in rdr:
            print("READ", command, format_flags, data, offset)
            try:
                if _chk(each_busy):
                    await rdr.write(0, format_flags, 0, data=None)
                await _schk(each_slow)
                if _chk(each_close):
                    return
                if command == OWMsg.nop:
                    await rdr.write(0, format_flags, 0)
                elif command == OWMsg.dirall:
                    data = data.rstrip(b"\0")
                    subtree = tree
                    path = []
                    for k in data.split(b"/"):
                        if k == b"":
                            continue
                        path.append(k)
                        k = k.decode("utf-8")
                        try:
                            subtree = subtree[k]
                        except KeyError:
                            raise NoEntryError(command, data)
                    res = []
                    for k in sorted(subtree.keys()):
                        k = k.encode("utf-8")
                        res.append(k)
                    if path:
                        path = b"/" + b"/".join(path) + b"/"
                    else:
                        path = b"/"
                    data = b",".join(path + k for k in res)
                    await rdr.write(0, format_flags, len(data), data + b"\0")
                elif command == OWMsg.read:
                    data = data.rstrip(b"\0")
                    res = tree
                    for k in data.split(b"/"):
                        if k == b"":
                            continue
                        k = k.decode("utf-8")
                        try:
                            res = res[k]
                        except KeyError:
                            raise NoEntryError(command, data)
                    if isinstance(res, dict):
                        raise IsDirError(command, data)
                    if not isinstance(res, bytes):
                        res = str(res).encode("utf-8")
                    await rdr.write(0, format_flags, len(res), res + b"\0")
                elif command == OWMsg.write:
                    val = data[-offset:].decode("utf-8")
                    data = data[:-offset]
                    data = data.rstrip(b"\0")
                    res = tree
                    last = None
                    for k in data.split(b"/"):
                        if k == b"":
                            continue
                        if last is not None:
                            try:
                                res = res[last]
                            except KeyError:
                                raise NoEntryError(command, data)
                        last = k.decode("utf-8")
                    assert last is not None
                    if last not in res:
                        raise NoEntryError(command, data)
                    res[last] = val
                    await rdr.write(0, format_flags, 0)
                else:
                    raise RuntimeError(
                        "Unknown command: %d %x %s %d"
                        % (command, format_flags, repr(data), offset)
                    )
            except OWFSReplyError as err:
                logger.info("Error: %s", err)
                await rdr.write(
                    -err.err, format_flags  # pylint: disable=invalid-unary-operand-type
                )

    except anyio.ClosedResourceError:
        pass
    finally:
        await socket.aclose()
        logger.debug("END Server")


class EventChecker:
    """
    This class is used to verify that whatever happens on the bus is what
    you'd expect to happen.

    Useful mainly as regression test, to ensure that new versions of your
    code actually do the requests they're supposed to do / don't generate
    more bus traffic.
    """

    pos: int = None

    def __init__(self, events=[]):  # pylint: disable=dangerous-default-value
        self.events = events[:]

    def add(self, event):
        self.events.push(event)

    async def __call__(self, ow, evt=None):
        self.pos = 0
        try:
            async with ow.events as ev:
                if evt is not None:
                    evt.set()
                async for e in ev:
                    if e is None:
                        break
                    logger.debug("Event %s", e)
                    self.check_next(e)
        except RuntimeError:
            raise
        else:
            self.check_last()
        # don't check on BaseException =Cancellation

    def check_next(self, e):
        try:
            t = self.events[self.pos]
        except IndexError:
            raise RuntimeError("Unexpected event %s" % (e,))
        self.pos += 1
        if isinstance(t, type) and isinstance(e, t):
            pass
        elif not (t == e):
            raise RuntimeError("Wrong event: want %s but has %s" % (t, e))
            # logger.error("Wrong event: want %s but has %s", t, e)

    def check_last(self):
        logger.debug("Event END")
        if self.pos != len(self.events):
            raise RuntimeError("Superfluous event #%d: %s" % (self.pos, self.events[self.pos]))


@asynccontextmanager
async def server(  # pylint: disable=dangerous-default-value  # intentional
    tree={}, options={}, events=None, polling=False, scan=None, initial_scan=True, **kw
):
    """
    This is a mock 1wire server+client.

    The context manager returns the client.

    ``tree`` and ``opotions`` are used as in `some_server`, ``polling``,
    ``scan`` and ``initial_scan`` are used to set up the client, other
    keyword arguments are forwarded to the client constructor.
    """
    PORT = (os.getpid() % 9999) + 40000
    async with OWFS(**kw) as ow:
        async with anyio.create_task_group() as tg:
            s = None
            listener = None
            try:
                listener = await anyio.create_tcp_listener(
                    local_host="127.0.0.1", local_port=PORT, reuse_port=True
                )

                async def may_close():
                    try:
                        with ungroup:
                            await listener.serve(partial(some_server, tree, options))
                    except (anyio.ClosedResourceError, anyio.BrokenResourceError):
                        pass

                if events is not None:
                    evt = anyio.Event()
                    tg.start_soon(events, ow, evt)
                    await evt.wait()
                addr = listener.extra(anyio.abc.SocketAttribute.raw_socket).getsockname()
                tg.start_soon(may_close)

                s = await ow.add_server(
                    *addr, polling=polling, scan=scan, initial_scan=initial_scan
                )
                ow.test_server = s
                yield ow
            finally:
                ow.test_server = None
                if listener is not None:
                    await listener.aclose()
                with anyio.CancelScope(shield=True):
                    if s is not None:
                        await s.drop()
                    await ow.push_event(None)
                    await anyio.sleep(0.1)
                tg.cancel_scope.cancel()



from setuptools import setup, find_packages

LONG_DESC = """\
This library allows you to access 1wire buses with Python, using any
async library (trio, asyncio, curio).

This code does *not* try to talk to any 1wire interface by itself; it
uses TCP port 4304 (by default) to access one or more ``owserver``\\ s.

Bus enumeration (when starting up, plus periodically in the background
if desired) is automatic and can be turned off.

This code supports background polling, alerts, and simultaneous
conversions. Bus couplers (DS2509) are handled transparently. Accessors to
device attributes are created from device descriptors, read from
``owserver`` and augmented by device-specific classes when necessary.

The name of this library is (a) historic, (b) its main author strongly
prefers Trio.
"""

setup(
    name="asyncowfs",
    use_scm_version={"version_scheme": "guess-next-dev", "local_scheme": "dirty-tag"},
    setup_requires=["setuptools_scm"],
    description="Async access to 1wire (asyncio,trio,curio)",
    url="https://github.com/smurfix/asyncowfs",
    long_description=LONG_DESC,
    author="Matthias Urlichs",
    author_email="matthias@urlichs.de",
    license="MIT -or- Apache License 2.0",
    packages=find_packages(),
    install_requires=["anyio >=4"],
    keywords=["trio", "async", "io", "networking"],
    python_requires=">=3.5.3",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Framework :: AsyncIO",
        "Framework :: Trio",
    ],
)



"""
Devices.
"""

import attr
import anyio
from typing import List

from .event import DeviceLocated, DeviceNotFound, DeviceValue, DeviceException
from .error import IsDirError

import logging

logger = logging.getLogger(__name__)

__all__ = ["Device"]


@attr.s
class NoLocationKnown(RuntimeError):
    device = attr.ib()


@attr.s
class NotADevice(RuntimeError):
    id = attr.ib()


dev_classes = dict()


def register(cls):
    dev_classes[cls.family] = cls


def split_id(id):  # pylint: disable=redefined-builtin
    try:
        a, b, c = (int(x, 16) for x in id.split("."))
    except ValueError:
        raise NotADevice(id) from None
    return a, b, c


class _Value:
    def __init__(self, path, typ):
        self.path = path
        self.typ = typ

    def __repr__(self):
        return "<%s: %s %s>" % (self.__class__.__name__, self.path, self.typ)


class _RValue(_Value):
    def __init__(self, path, typ):
        super().__init__(path, typ)
        if typ in {"f", "g", "p", "t"}:
            self.conv = float
        elif typ in {"i", "u"}:
            self.conv = int
        elif typ == "y":
            self.conv = lambda x: bool(int(x))
        elif typ == "b":
            self.conv = lambda x: x
        else:
            self.conv = lambda x: x.decode("utf-8")


class _WValue(_Value):
    def __init__(self, path, typ):
        super().__init__(path, typ)
        if typ == "b":
            self.conv = lambda x: x
        elif typ == "y":
            self.conv = lambda x: b"1" if x else b"0"
        else:
            self.conv = lambda x: str(x).encode("utf-8")


class SimpleValue(_RValue):
    """Accessor for direct attribute access"""

    async def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        res = await self.dev.attr_get(*slf.path)
        return slf.conv(res)


class SimpleGetter(_RValue):
    """Accessor for get_* function"""

    def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        async def getter():
            res = await self.dev.attr_get(*slf.path)
            return slf.conv(res)

        return getter


class SimpleSetter(_WValue):
    """Accessor for set_* function"""

    def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        async def setter(val):
            await self.dev.attr_set(*slf.path, value=slf.conv(val))

        return setter


class MultiValue(_RValue):
    """Accessor for direct array access"""

    async def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        p = slf.path[:-1] + (slf.path[-1] + ".ALL",)
        res = await self.dev.attr_get(*p)
        conv = slf.conv
        return [conv(v) for v in res.split(b",")]


class MultiGetter(_RValue):
    """Accessor for array get_* function"""

    def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        async def getter():
            p = slf.path[:-1] + (slf.path[-1] + ".ALL",)
            res = await self.dev.attr_get(*p)
            conv = slf.conv
            return [conv(v) for v in res.split(b",")]

        return getter


class MultiSetter(_WValue):
    """Accessor for array set_* function"""

    def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        async def setter(val):
            conv = slf.conv
            p = slf.path[:-1] + (slf.path[-1] + ".ALL",)
            val = b",".join(conv(v) for v in val)
            await self.dev.attr_set(*p, value=val)

        return setter


class _IdxObj:
    def __init__(self, dev, ary):
        self.dev = dev
        self.ary = ary

    async def __getitem__(self, idx):
        if self.ary.num:
            idx = str(idx)
        else:
            idx = chr(ord("A") + idx)
        p = self.ary.path[:-1] + (self.ary.path[-1] + "." + idx,)
        res = await self.dev.attr_get(*p)
        return self.ary.conv(res)

    async def set(self, idx, val):
        if self.ary.num:
            idx = str(idx)
        else:
            idx = chr(ord("A") + idx)
        p = self.ary.path[:-1] + (self.ary.path[-1] + "." + idx,)
        await self.dev.attr_set(*p, value=val)


class ArrayValue(_RValue):
    """Accessor for direct array element access"""

    def __init__(self, path, typ, num):
        super().__init__(path, typ)
        self.num = num

    def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        return _IdxObj(self.dev, slf)


class ArrayGetter(_RValue):
    """Accessor for array element get_* function"""

    def __init__(self, path, typ, num):
        super().__init__(path, typ)
        self.num = num

    def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        async def getter(idx):
            if slf.num:
                idx = str(idx)
            else:
                idx = chr(ord("A") + idx)
            p = slf.path[:-1] + (slf.path[-1] + "." + idx,)
            res = await self.dev.attr_get(*p)
            return slf.conv(res)

        return getter


class ArraySetter(_WValue):
    """Accessor for array element set_* function"""

    def __init__(self, path, typ, num):
        super().__init__(path, typ)
        self.num = num

    def __get__(slf, self, cls):  # pylint: disable=no-self-argument
        async def setter(idx, val):
            if slf.num:
                idx = str(idx)
            else:
                idx = chr(ord("A") + idx)
            p = slf.path[:-1] + (slf.path[-1] + "." + idx,)
            await self.dev.attr_set(*p, value=slf.conv(val))

        return setter


class SubDir:
    _subdirs = set()
    dev = None  # needs to be filled by subclass

    def __getattr__(self, name):
        if name not in self._subdirs:
            return super().__getattribute__(name)
        c = getattr(self, "_cls_" + name)(self)
        c.dev = self.dev
        return c


async def setup_accessors(server, cls, typ, *subdir):
    cls.fields = {}
    for d in await server.dir("structure", typ, *subdir):
        dd = subdir + (d,)
        try:
            v = await server.attr_get("structure", typ, *dd)
        except IsDirError:

            t = typ

            class SubPath(SubDir):
                typ = t
                subdir = dd

                def __init__(self, base):
                    self.base = base

                def __repr__(self):
                    return "<%s %s %s>" % (
                        self.__class__.__name__,
                        self.base,
                        self.subdir,
                    )

                def __get__(self, obj, cls):
                    if obj is None:
                        return cls
                    try:
                        return getattr(obj, "_" + self.dd[-1])
                    except AttributeError:
                        c = getattr(cls, self.__name__)()
                        setattr(obj, "_" + self.dd[-1], c)
                        c.dev = obj.dev
                        return c

            SubPath.__name__ = "_cls_" + d
            setattr(cls, "_cls_" + d, SubPath)
            cls._subdirs.add(d)
            await setup_accessors(server, SubPath, typ, *dd)

        else:
            v = v.decode("utf-8").split(",")
            try:
                v[1] = int(v[1])
                v[2] = int(v[2])
                v[4] = int(v[4])
            except ValueError:
                raise ValueError("broken setup vector", (typ, dd), v) from None
            if v[1] == 0:
                if d.endswith(".0"):
                    num = True
                elif d.endswith(".A"):
                    num = False
                else:
                    num = None
                v[1] = num
                cls.fields[d] = v

                if num is None:
                    if v[3] in {"ro", "rw"}:
                        if hasattr(cls, d):
                            logger.debug("%s: not overwriting %s", cls, d)
                        else:
                            setattr(cls, d, SimpleValue(dd, v[0]))
                        setattr(cls, "get_" + d, SimpleGetter(dd, v[0]))
                    if v[3] in {"wo", "rw"}:
                        setattr(cls, "set_" + d, SimpleSetter(dd, v[0]))
                else:
                    d = d[:-2]
                    dd = subdir + (d,)
                    if v[3] in {"ro", "rw"}:
                        if hasattr(cls, d):
                            logger.debug("%s: not overwriting %s", cls, d)
                        else:
                            setattr(cls, d, ArrayValue(dd, v[0], num))
                        setattr(cls, "get_" + d, ArrayGetter(dd, v[0], num))
                    if v[3] in {"wo", "rw"}:
                        setattr(cls, "set_" + d, ArraySetter(dd, v[0], num))

                    if v[3] in {"ro", "rw"}:
                        if hasattr(cls, d + "_all"):
                            logger.debug("%s: not overwriting %s", cls, d + "_all")
                        else:
                            setattr(cls, d + "_all", MultiValue(dd, v[0]))
                        setattr(cls, "get_" + d + "_all", MultiGetter(dd, v[0]))
                    if v[3] in {"wo", "rw"}:
                        setattr(cls, "set_" + d + "_all", MultiSetter(dd, v[0]))


class Device(SubDir):
    """Base class for devices.

    A device may or may not have a known location.

    Whenever a device is located, poll activity is auto-started.
    """

    _did_setup = False
    _poll: dict = None
    bus = None
    _events = None
    _wait_bus = None

    def __new__(cls, service, id):  # pylint: disable=redefined-builtin
        family_id, code, chksum = split_id(id)

        cls = dev_classes.get(family_id)  # pylint: disable=self-cls-assignment
        if cls is None:

            class cls(Device):  # pylint: disable=function-redefined
                family = family_id

            cls.__name__ = "Device_%02x" % (family_id,)
            dev_classes[family_id] = cls

        self = object.__new__(cls)

        self.id = id.upper()
        self.family = family_id
        self.code = code
        self.chksum = chksum

        self.service = service
        self.bus = None

        self._unseen = 0
        self._events = []
        self._wait_bus = anyio.Event()
        self._poll = {}  # name > poll task scopes
        self._intervals = {}
        self._task_lock = anyio.Lock()

        return self

    def __init__(self, service, id):  # pylint: disable=redefined-builtin,unused-argument
        logger.debug("NewDev %s", id)

    @property
    def dev(self):
        return self

    def queue_event(self, evt):
        """Remember this event. Used if an event arrives when the device
        hasn't yet been set up by high-level code
        """
        self._events.append(evt)

    @property
    def queued_events(self):
        """Return queued events. Shall be called exactly once."""
        e, self._events = self._events, None
        if e is None:
            raise RuntimeError("You cannot call `queued_events` " "more than once")
        return iter(e)

    @classmethod
    async def setup_struct(cls, server):
        """Read the device's structural data from OWFS
        and add methods to access the fields"""

        if cls._did_setup is not False:
            return
        cls._did_setup = None

        try:
            fc = "%02X" % (cls.family)
            await setup_accessors(server, cls, fc)

        except BaseException:
            cls._did_setup = False
            raise
        else:
            cls._did_setup = True

    def __eq__(self, x):
        x = getattr(x, "id", x)
        return self.id == x

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return "<%s:%s @ %s>" % (self.__class__.__name__, self.id, self.bus)

    def buses(self):
        return set()

    async def locate(self, bus):
        """The device has been seen here."""
        if self.bus is bus:
            return
        self.bus = bus
        self._wait_bus.set()
        await self.service.push_event(DeviceLocated(self))
        for typ, val in self._intervals.items():
            await self._set_poll_task(typ, val)

    async def wait_bus(self):
        await self._wait_bus.wait()

    async def delocate(self, bus):
        """The device is no longer located here."""
        if self.bus is bus:
            self._wait_bus = anyio.Event()
            await self._delocate()

    async def _delocate(self):
        await self.bus._del_device(self)
        self.bus = None
        for t in self._poll.values():
            t.cancel()
        self._poll = {}
        await self.service.push_event(DeviceNotFound(self))

    async def attr_get(self, *attrs: List[str]):
        """Read this attribute (ignoring device struct)"""
        if self.bus is None:
            raise NoLocationKnown(self)
        return await self.bus.attr_get(self.id, *attrs)

    async def attr_set(self, *attrs: List[str], value):
        """Write this attribute (ignoring device struct)"""
        if self.bus is None:
            raise NoLocationKnown(self)
        return await self.bus.attr_set(self.id, *attrs, value=value)

    async def get(self, *attrs):
        """Read this attribute (following device struct)"""
        dev = self
        for k in attrs:
            if isinstance(k, int):
                dev = dev[k]
            else:
                dev = getattr(dev, k)
        return await dev

    async def set(self, *attrs, value):
        """Write this attribute (following device struct)"""
        dev = self
        for k in attrs[:-1]:
            if isinstance(k, int):
                dev = dev[k]
            else:
                dev = getattr(dev, k)
        if isinstance(dev, _IdxObj):
            await dev.set(attrs[-1], value)
        else:
            await getattr(dev, "set_" + attrs[-1])(value)

    def polling_items(self):
        """Enumerate poll variants supported by this device.

        This is a generator. If you override, call::

            yield from super().polling_items()

        See the associated ``poll_<name>`` methods on
        :class:`anyio_owfs.bus.Bus` for details.

        Special return values:

        * "alarm": you need to implement ``.stop_alarm``
        """
        if False:  # pylint: disable=using-constant-test
            yield None

    def polling_interval(self, typ: str):
        """Return the interval WRT how often to poll for this type.

        The default implementation looks up the "interval_<typ>" attribute
        or returns ``None`` if that doesn't exist.
        """
        try:
            return self._intervals[typ]
        except KeyError:
            return getattr(self, "interval_" + typ, None)

    async def set_polling_interval(self, typ: str, value: float = 0):
        if isinstance(typ, str):
            styp = typ
        else:
            styp = "/".join(str(x) for x in typ)
        if value > 0:
            self._intervals[styp] = value
        else:
            self._intervals.pop(styp, None)

        if self.bus is not None:
            if hasattr(self, "poll_" + styp):
                await self.bus.update_poll()
            else:
                await self._set_poll_task(typ, value)

    async def _set_poll_task(self, typ, value):
        async with self._task_lock:
            try:
                task = self._poll.pop(typ)
            except KeyError:
                pass
            else:
                task.cancel()
            if not value:
                return

            *p, n = typ.split("/") if isinstance(typ, str) else typ
            s = self
            for pp in p:
                if isinstance(pp, int):
                    s = s[pp]
                else:
                    s = getattr(s, pp)
            if isinstance(n, int) or hasattr(s, "get_" + n):
                self._poll[typ] = await self.service.add_task(self._poll_task, s, n, typ, value)

            else:
                raise RuntimeError("%r: No poll for %s" % (self, typ))

    async def _poll_task(self, s, n, typ, value):
        await anyio.sleep(value / 5)
        while True:
            try:
                if isinstance(n, int):
                    v = await s[n]
                else:
                    v = await getattr(s, n)
            except Exception as exc:
                logger.error("Reader at %s %s: %r", self, typ, exc)
                await self.service.push_event(DeviceException(self, typ, exc))
            else:
                await self.service.push_event(DeviceValue(self, typ, v))
            await anyio.sleep(value)

    async def poll_alarm(self):
        """Tells the device not to trigger an alarm any more.

        You *need* to override this if your device can trigger an alarm
        condition. Also, this method *must* disable the alarm; your
        application can re-enable it later, when processing the
        :class:`anyio_owfs.event.DeviceAlarm` event.
        """
        pass  # pylint
        raise NotImplementedError(
            "<%s> (%02x) needs 'poll_alarm'" % (self.__class__.__name__, self.family)
        )


@register
class SwitchDevice(Device):
    family = 0x1F

    def buses(self):
        """Return a list of the connected buses.
        'main' should be processed first.
        """
        b = []
        b.append((self.id, "main"))
        b.append((self.id, "aux"))
        return b

    async def poll_alarm(self):
        """Clear alarm"""
        raise RuntimeError("TODO")
        # res = await self.get_event_all()
        # await self.set_clearalarm(1)


@register
class TemperatureDevice(Device):
    family = 0x10

    interval_temperature = None
    alarm_temperature = None

    async def poll_alarm(self):
        """Turn off alarm condition by adapting the temperature bounds"""
        self.alarm_temperature = t = await self.latesttemp
        reasons = {"temp": t}

        t_h = await self.temphigh
        if t > t_h:
            await self.set_temphigh(int(t + 2))
            reasons["high"] = t_h
        t_l = await self.templow
        if t < t_l:
            await self.set_templow(int(t - 1))
            reasons["low"] = t_l
        return reasons

    def polling_items(self):
        yield from super().polling_items()
        yield "temperature"
        yield "alarm"

    async def poll_temperature(self, simul=False):
        # Bug workaround
        t = await (self.latesttemp if simul and len(self.bus.path)<=1 else self.temperature)
        if t == 85:
            logger.error("TEMP: got 85 on %r", self)
            return

        await self.service.push_event(DeviceValue(self, "temperature", t))


#   @property
#   def temperature(self):
#       return self.latesttemp


@register
class TemperatureBDevice(Device):
    family = 0x28


@register
class VoltageDevice(Device):
    family = 0x20

    interval_voltage = None
    alarm_voltage = None

    async def poll_alarm(self):
        """Turn off alarm condition by adapting the voltage bounds"""
        reasons = {}
        v = await self.voltage_all
        ah = await self.alarm.high_all
        al = await self.alarm.low_all
        power = await self.set_alarm.unset
        if power:
            reasons["power_on"] = True
            await self.set_alarm.set_unset(0)
        if any(ah):
            vh = await self.set_alarm.set_high_all
            for i in range(4):
                if not ah[i]:
                    continue
                reasons["high_" + str(i)] = vh[i]
                await self.set_alarm.set_high(i, 0)
        if any(al):
            vl = await self.set_alarm.set_low_all
            for i in range(4):
                if not al[i]:
                    continue
                reasons["low_" + str(i)] = vl[i]
                await self.set_alarm.set_low(i, 0)
        for i in range(4):
            reasons["volt_" + str(i)] = v[i]

        return reasons

    def polling_items(self):
        yield from super().polling_items()
        yield "voltage"
        yield "alarm"

    async def poll_voltage(self, simul=False):
        v = await self.volt_all
        await self.service.push_event(DeviceValue(self, "volt_all", v))


@register
class PIODevice(Device):
    family = 0x05
    # dumb slave, no special handling



"""set.py -- an AsyncOWFS implementation of owget

Usage examples:

python3 set.py 10.67C6697351FF.BF temphigh 50

"""

from __future__ import print_function

import sys

import collections

import asyncclick as click
from asyncowfs import OWFS

import logging

logger = logging.getLogger("examples.walk")

__all__ = ["main"]


async def mon(ow):
    async with ow.events as events:
        async for msg in events:
            logger.info("%s", msg)


@click.command()
@click.option("--host", "-h", default="localhost", help="host running owserver")
@click.option("--port", "-p", default=4304, type=int, help="owserver port")
@click.option("--debug", "-D", is_flag=True, help="Show debug information")
@click.argument("id")
@click.argument("attr")
@click.argument("data")
async def main(host, port, debug, id, attr, data):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    async with OWFS() as ow:
        if debug:
            await ow.add_task(mon, ow)
        s = await ow.add_server(host, port)
        attr = [k for k in attr.split("/") if k]
        dev = await ow.get_device(id)
        if dev.bus is None:
            print("Device not found", file=sys.stderr)
            sys.exit(1)
        await dev.attr_set(*attr, value=data)


if __name__ == "__main__":
    main()



# utility code

import attr
import outcome

import anyio
from concurrent.futures import CancelledError


@attr.s
class ValueEvent:
    """A waitable value useful for inter-task synchronization,
    inspired by :class:`threading.Event`.

    An event object manages an internal value, which is initially
    unset, and tasks can wait for it to become True.

    """

    event = attr.ib(factory=anyio.Event, init=False)
    value = attr.ib(default=None, init=False)

    def set(self, value):
        """Set the internal flag value to True, and wake any waiting tasks."""
        self.value = outcome.Value(value)
        self.event.set()

    def is_set(self):
        return self.value is not None

    def set_error(self, exc):
        """Set the internal flag value to True, and wake any waiting tasks."""
        self.value = outcome.Error(exc)
        self.event.set()

    def cancel(self):
        self.set_error(CancelledError())

    async def get(self):
        """Block until the internal flag value becomes True.

        If it's already True, then this method is still a checkpoint, but
        otherwise returns immediately.

        """
        await self.event.wait()
        return self.value.unwrap()
